/* --------------------------------------------------------------------------
 * core of haserl.cgi - a poor-man's php for embedded/lightweight environments
 * $Id: haserl.c,v 1.25 2005/11/03 16:51:20 nangel Exp $ 
 * Copyright (c) 2003,2004    Nathan Angelacos (nangel@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * -----
 * The x2c() and unescape_url() routines were taken from  
 *  http://www.jmarshall.com/easy/cgi/getcgi.c.txt 
 * 
 * The comments in that text file state:
 *
 ***  Written in 1996 by James Marshall, james@jmarshall.com, except 
 ***  that the x2c() and unescape_url() routines were lifted directly 
 ***  from NCSA's sample program util.c, packaged with their HTTPD. 
 ***     For the latest, see http://www.jmarshall.com/easy/cgi/ 
 * -----
 *
 ------------------------------------------------------------------------- */
/*Splint info follows:  */
/* splint +posixlib webconf.c */
/*@-mustfreeonly -branchstate@*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <getopt.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <stdlib.h>
#include <string.h>

#if HAVE_SIGNAL_H
#include <signal.h>
#endif

#include "haserl.h"
#include "common.h"


const char *REVTAG = "$Id: haserl.c,v 1.25 2005/11/03 16:51:20 nangel Exp $";

#ifndef TEMPDIR
#define TEMPDIR "/tmp"
#endif

#ifndef MAX_UPLOAD_KB
#define MAX_UPLOAD_KB 2048
#endif

/* Refuse to disable the subshell */
#ifndef SUBSHELL_CMD
#define SUBSHELL_CMD "/bin/sh"
#endif

buflist_t /*@null@ */  * env_list = NULL;
token_t /*@null@ */  * token_list = NULL;

char global_variable_prefix[] = HASERL_VAR_PREFIX;
int global_subshell_pipe[4];
int global_subshell_pid;
int global_subshell_died = 0;

/* Just a silly construct to contain global variables    */
static struct haserl_t
{
  unsigned long uploadkb;	/* how big an upload do we allow (0 for none) */
  char *shell;			/* The shell we use                                     */
  char *uploaddir;		/* where we upload to                                   */
  int verbatim;			/* true if we print whitespace html segments (old-style) */
  int debug;			/* true if in "debug" mode                              */
  int bangscript;		/* true if run as a #!.... script                       */
  int acceptall;		/* true if we'll accept POST data on GETs and vice versa */
} global;



/*
 * Command line / Config file directives When adding a long option, make sure
 * to update the short_options as well
 */

struct option ga_long_options[] = {
  {"version", no_argument, 0, 'v'},
  {"help", no_argument, 0, 'h'},
  {"debug", no_argument, 0, 'D'},
  {"upload-limit", optional_argument, 0, 'u'},
  {"upload-dir", required_argument, 0, 'U'},
  {"accept-all", no_argument, 0, 'a'},
  {"shell", required_argument, 0, 's'},
  {"verbatim", no_argument, 0, 'V'},
  {0, 0, 0, 0}
};

const char *gs_short_options = "vhDu::U:as:V";

/*
 * enum tag_t { HTML, RUN, IF, ELSE, FI, ABORT, INCLUDE,  NOOP };
 */


enum h_error_t
{ NO_ERROR, LT, GT, MALLOC, MISSING_X, OPENFAIL, X_BEFORE_Y };

const char *global_error[] = {
  "",				/* NO_ERROR             */
  "&lt;?",			/* LT                */
  "?&gt;",			/* GT                */
  "Memory Allocation Error",	/* STACK_UNDERRUN       */
  "Missing %s in line %d",	/* MISSING_X */
  "Unable to open file %s in line %d",	/* OPENFAIL */
  "Found %s before %s near line %d",	/* X_BEFORE_Y   */
  ""
};


const char *global_tag[] = {
  "",
  "",
  "if",
  "el",
  "fi",
  "ab",
  "in",
  ""
};


/*
 * Convert 2 char hex string into char it represents
 * (from http://www.jmarshall.com/easy/cgi)
 */
char
x2c (char *what)
{
  char digit;

  /*@+charint@ */
  digit = (what[0] >= 'A' ? ((what[0] & 0xdf) - 'A') + 10 : (what[0] - '0'));
  digit *= 16;
  digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A') + 10 : (what[1] - '0'));
  return (digit);
  /*@-charint@ */
}

/*
 * unsescape %xx to the characters they represent 
 */

void
unescape_url (char *url)
{
  int i, j;
  /*@-predboolothers@ */
  for (i = 0, j = 0; url[j]; ++i, ++j)
    {
      if ((url[i] = url[j]) == '%')
	{
	  url[i] = x2c (&url[j + 1]);
	  j += 2;
	}
    }
  url[i] = '\0';
  /*@+predboolothers@ */
}

/*
 * myputenv (char *newbuf, char *prefix )
 *   Searches for an existing "key=" value in the chain;
 *   if one exists, it removes it from the environment, deletes the 
 *   existing link, then adds the new varaiable in
 *   Returns -1 on error, 0 on success
 *
 *   prefix is appended to the environment string (e.g. US_blah=foo )
 */
int
myputenv (char *newbuf, char *prefix)
{
  buflist_t *cur = env_list;
  buflist_t *prev = NULL;
  size_t keylen;
  char *entry = NULL;

  entry = calloc (sizeof (char), strlen (newbuf) + strlen (prefix) + 1);
  if (entry == NULL)
    {
      return (-1);
    }
  if (strlen (prefix))
    {
      memcpy (entry, prefix, strlen (prefix));
    }
  memcpy ((char *) (entry + strlen (prefix)), newbuf, strlen (newbuf));

  keylen = (size_t) (index (entry, '=') - entry);

  if (keylen <= 0)
    {
      free (entry);
      return (-1);
    }

  /* does the value already exist? */
  while (cur != NULL)
    {
      if (memcmp (cur->buf, entry, keylen + 1) == 0)
	{
	  entry[keylen] = '\0';
	  unsetenv (entry);
	  entry[keylen] = '=';
	  free (cur->buf);
	  if (prev != NULL)
	    prev->next = cur->next;
	  free (cur);
	  cur = prev;
	}			/* end if found a matching key */
      prev = cur;
      cur = (buflist_t *) cur->next;
    }				/* end if matching key */

  /* add the value to the end of the chain  */
  if (cur == NULL)
    {				/* paranoia... cur should always be null */
      cur = malloc (sizeof (buflist_t));
      cur->buf = entry;
      putenv (cur->buf);
      if (prev != NULL)
	prev->next = cur;
    }
  else
    {
      return (-1);
    }

  return (0);
}


/* readenv 
 * reads the current environment an popluates our environment chain 
 */

void
readenv ()
{
  extern char **environ;
  int count = 0;

  while (environ[count] != NULL)
    {
      myputenv (environ[count], "");
      count++;
    }
}


 /* freeenvchain
  * Misnamed, but basically goes through the linked list of 
  * buflist and clears the entire chain, but does not clear 
  * the environment
  */

void
freeenvchain ()
{
  buflist_t *cur = env_list;
  buflist_t *next;
  while (cur != NULL)
    {
      free (cur->buf);
      next = (buflist_t *) cur->next;
      free (cur);
      cur = next;
    }

}

/* CookieVars ()
 * if HTTP_COOKIE is passed as an environment variable,
 * attempt to parse its values into environment variables
 */
void
CookieVars ()
{
  char *qs;
  char *token;

  if (getenv ("HTTP_COOKIE") != NULL)
    {
      qs = strdup (getenv ("HTTP_COOKIE"));
    }
  else
    {
      return;
    }

	/** split on; to extract name value pairs */
  token = strtok (qs, ";");
  while (token)
    {
      // skip leading spaces 
      while (token[0] == ' ')
	{
	  token++;
	}
      myputenv (token, global_variable_prefix);
      token = strtok (NULL, ";");
    }
  free (qs);
}


/* FreeTokenChain () 
 * goes through and frees the token chain 
 */

void
FreeTokenChain ()
{
  token_t *cur = token_list;
  token_t *next;
  while (cur != NULL)
    {
      next = (token_t *) cur->next;
      free (cur);
      cur = next;
    }
}


/* SessionID 
 *  Makes a uniqe SESSIONID environment variable for this script 
 */

void
sessionid ()
{
  char session[29];

  sprintf (session, "SESSIONID=%x%x", getpid (), (int) time (NULL));
  myputenv (session, "");
}

void
wcversion ()
{
  char version[200];
  sprintf (version, "HASERLVER=%s", PACKAGE_VERSION);
  myputenv (version, "");
}


void
haserlflags ()
{
  char buf[200];

  snprintf (buf, 200, "HASERL_UPLOAD_DIR=%s", global.uploaddir);
  myputenv (buf, "");

  snprintf (buf, 200, "HASERL_UPLOAD_LIMIT=%lu", global.uploadkb);
  myputenv (buf, "");

  snprintf (buf, 200, "HASERL_ACCEPT_ALL=%d", global.acceptall);
  myputenv (buf, "");

  snprintf (buf, 200, "HASERL_SHELL=%s", global.shell);
  myputenv (buf, "");

  snprintf (buf, 200, "HASERL_VERBATIM=%d", global.verbatim);
  myputenv (buf, "");

  snprintf (buf, 200, "HASERL_CGIMODE=%d", global.bangscript);
  myputenv (buf, "");

}

/* 
 * Read cgi variables from query string, and put in environment
 */

int
ReadCGIQueryString ()
{
  char *qs;
  char *token;
  int i;

  if (getenv ("QUERY_STRING") != NULL)
    {
      qs = strdup (getenv ("QUERY_STRING"));
    }
  else
    {
      return (0);
    }

  /* change plusses into spaces */
  for (i = 0; qs[i]; i++)
    {
      if (qs[i] == '+')
	{
	  qs[i] = ' ';
	}
    };

	/** split on & and ; to extract name value pairs */

  token = strtok (qs, "&;");
  while (token)
    {
      unescape_url (token);
      myputenv (token, global_variable_prefix);
      token = strtok (NULL, "&;");
    }
  free (qs);
  return (0);
}


/* 
 * Read cgi variables from stdin (for POST queries)
 * (oh... and if its mime-encoded file upload, we save the
 * file to /tmp; and return the name of the tmp file
 * the cgi script is responsible for disposing of the tmp file
 */

int
ReadCGIPOSTValues ()
{
  char *qs;
  int content_length;
  int i;
  char *token;


  if (getenv ("CONTENT_LENGTH") == NULL)
    {
      /* no content length?  Trash this request! */
      PrintParseError ("HTTP POST request did not specify a Content Length.",
		       0);
      return (-1);
    }
  else
    {
      content_length = atoi (getenv ("CONTENT_LENGTH"));
    }

  /* Allow 2MB content, unless they have a global upload was set */
  if (content_length > (((global.uploadkb == 0)?2048:global.uploadkb) * 1024))
    {
      PrintParseError
	("Attempted to send content larger than allowed limits.", 0);
      /* But we need to finish reading the content */
      while (fread (&i, sizeof (int), 1, stdin) == 1);
      return (-1);
    }

  if (!(qs = malloc (content_length + 1)))
    {
      PrintParseError ("Unable to Allocate memory for POST content.", 0);
      return (-1);
    }

  /* set the buffer to null, so that a browser messing with less 
     data than content_length won't buffer underrun us */
  memset (qs, 0, content_length + 1);

  if ((!fread (qs, content_length, 1, stdin) && (content_length > 0)
       && !feof (stdin)))
    {
      PrintParseError ("Unable to read from stdin", 0);
      printf ("Content Length is %d ", content_length);
      free (qs);
      return (-1);
    }
  if (getenv ("CONTENT_TYPE"))
    {
      if (strncasecmp (getenv ("CONTENT_TYPE"), "multipart/form-data", 19) ==
	  0)
	{
	  /* This is a mime request, we need to go to the mime handler */
	  i = ReadMimeEncodedInput (qs);
	  free (qs);
	  return (i);
	}
    }

  /* change plusses into spaces */
  for (i = 0; qs[i]; i++)
    {
      if (qs[i] == '+')
	{
	  qs[i] = ' ';
	}
    };

	/** split on & and ; to extract name value pairs */

  token = strtok (qs, "&;");
  while (token)
    {
      unescape_url (token);
      myputenv (token, global_variable_prefix);
      token = strtok (NULL, "&;");
    }
  free (qs);
  return (0);
}

/*
 *  LineToStr - Scans char and replaces the first "\n" with a "\0";
 *  If it finds a "\r", it does that to; (fix DOS brokennes) returns
 *  the length of the string;
 */
int
LineToStr (char *string, size_t max)
{
  size_t offset = 0;

  while ((offset < max) && (string[offset] != '\n')
	 && (string[offset] != '\r'))
    {
      offset++;
    }
  if (string[offset] == '\r')
    {
      string[offset] = '\0';
      offset++;
    }
  if (string[offset] == '\n')
    {
      string[offset] = '\0';
      offset++;
    }
  return (offset);
}


/*
 * ReadMimeEncodedInput - handles things that are mime encoded
 * takes a pointer to the input; returns 0 on success
 */

int
ReadMimeEncodedInput (char *qs)
{
  char *boundary;
  char *ct;
  int i;
  int datastart;
  size_t cl;
  size_t offset;
  char *envname;
  char *filename;
  char *ptr;
  int line;
  char *tmpname;
  int fd;
  /* we should only get here if the content type was set. Segfaults happen
     if Content_Type is null */

  if (getenv ("CONTENT_LENGTH") == NULL)
    {
      /* No content length?! */
      PrintParseError ("No Content Length in HTTP Header from client", 0);
      return (-1);
    }

  cl = atoi (getenv ("CONTENT_LENGTH"));

  /* we do this 'cause we can't mess with the real env. variable - it would
   * overwrite the environment - I tried.
   */
  i = strlen (getenv ("CONTENT_TYPE")) + 1;
  ct = malloc (i);
  if (ct)
    {
      memcpy (ct, getenv ("CONTENT_TYPE"), i);
    }
  else
    {
      return (-1);
    }

  i = (int) NULL;
  if (ct != NULL)
    {
      while (i < strlen (ct) && (strncmp ("boundary=", &ct[i], 9) != 0))
	{
	  i++;
	}
    }
  if (i == strlen (ct))
    {
      /* no boundary informaiton found */
      PrintParseError ("No Mime Boundary Information Found", 0);
      free (ct);
      return (-1);
    }
  boundary = &ct[i + 7];
  /* add two leading -- to the boundary */
  boundary[0] = '-';
  boundary[1] = '-';

  /* begin the big loop.  Look for:
     --boundary
     Content-Disposition: form-data;  name="......." 
     ....
     <blank line>
     content
     --boundary
     Content-Disposition: form-data; name="....." filename="....."
     ...
     <blank line>
     --boundary--
     eof
   */

  offset = 0;
  while (offset < cl)
    {
      /* first look for boundary */
      while ((offset < cl)
	     && (memcmp (&qs[offset], boundary, strlen (boundary))))
	{
	  offset++;
	}
      /* if we got here and we ran off the end, its an error          */
      if (offset >= cl)
	{
	  PrintParseError ("Malformed MIME Encoding", 0);
	  free (ct);
	  return (-1);
	}
      /* if the two characters following the boundary are --,         */
      /* then we are at the end, exit                                 */
      if (memcmp (&qs[offset + strlen (boundary)], "--", 2) == 0)
	{
	  offset += 2;
	  break;
	}
      /* find where the offset should be */
      line = LineToStr (&qs[offset], cl - offset);
      offset += line;

      /* Now we're going to look for content-disposition              */
      line = LineToStr (&qs[offset], cl - offset);
      if (strncasecmp (&qs[offset], "Content-Disposition", 19) != 0)
	{
	  /* hmm... content disposition was not where we expected it */
	  PrintParseError ("Content-Disposition Missing", 0);
	  free (ct);
	  return (-1);
	}
      /* Found it, so let's go find "name="                           */
      if (!(envname = strstr (&qs[offset], "name=")))
	{
	  /* now name= is missing?!                               */
	  PrintParseError ("Content-Disposition missing name tag", 0);
	  free (ct);
	  return (-1);
	}
      else
	{
	  envname += 6;
	}
      /* is there a filename tag?                                     */
      if ((filename = strstr (&qs[offset], "filename=")) != NULL)
	{
	  filename += 10;
	}
      else
	{
	  filename = NULL;
	}
      /* make envname and filename ASCIIZ                             */
      i = 0;
      while ((envname[i] != '"') && (envname[i] != '\0'))
	{
	  i++;
	}
      envname[i] = '\0';
      if (filename)
	{
	  i = 0;
	  while ((filename[i] != '"') && (filename[i] != '\0'))
	    {
	      i++;
	    }
	  filename[i] = '\0';
	}
      offset += line;
      /* Ok, by some miracle, we have the name; let's skip till we    */
      /* come to a blank line                                         */
      line = LineToStr (&qs[offset], cl - offset);
      while (strlen (&qs[offset]) > 1)
	{
	  offset += line;
	  line = LineToStr (&qs[offset], cl - offset);
	}
      offset += line;
      datastart = offset;
      /* And we go back to looking for a boundary */
      while ((offset < cl)
	     && (memcmp (&qs[offset], boundary, strlen (boundary))))
	{
	  offset++;
	}
      /* strip [cr] lf */
      if ((qs[offset - 1] == '\n') && (qs[offset - 2] == '\r'))
	{
	  offset -= 2;
	}
      else
	{
	  offset -= 1;
	}
      qs[offset] = 0;
      /* ok, at this point, we know where the name is, and we know    */
      /* where the content is... we have to do one of two things      */
      /* based on whether its a file or not                           */
      if (filename == NULL)
	{			/* its not a file, so its easy            */
	  /* just jam the content after the name          */
	  memcpy (&envname[strlen (envname) + 1], &qs[datastart],
		  offset - datastart + 1);
	  envname[strlen (envname)] = '=';
	  myputenv (envname, global_variable_prefix);
	}
      else
	{			/* handle the fileupload case           */
	  if (offset - datastart)
	    {			/* only if they uploaded */
	      if (global.uploadkb == 0)
		{
		  PrintParseError ("File uploads not allowed here.", 0);
		  return (-1);
		}
	      /*  stuff in the filename */
	      ptr =
		calloc (sizeof (char),
			strlen (envname) + strlen (filename) + 2 + 5);
	      sprintf (ptr, "%s_name=%s", envname, filename);
	      myputenv (ptr, global_variable_prefix);
	      free (ptr);

	      tmpname = calloc (strlen (global.uploaddir) + 8, sizeof (char));
	      if (!tmpname)
		{
		  PrintParseError ("Memory Allocation Error", 0);
		  return (-1);
		}
	      strcpy (tmpname, global.uploaddir);
	      strcat (tmpname, "/XXXXXX");
	      fd = mkstemp (tmpname);
	      if (fd == -1)
		{
		  PrintParseError ("Unable to open temp file", 0);
		  return (-1);
		}
	      write (fd, &qs[datastart], offset - datastart);
	      close (fd);
	      ptr =
		calloc (sizeof (char),
			strlen (envname) + strlen (tmpname) + 2);
	      sprintf (ptr, "%s=%s", envname, tmpname);
	      myputenv (ptr, global_variable_prefix);
	      free (ptr);
	      free (tmpname);
	    }
	}
    }
  free (ct);
  return (0);
}

/* 
 * insertScript - Attempts to open a file and insert it into an existing
 * script buffer after script->curpos
 * returns 0 on success
 * 		-1 on failure
 */
int
insertScript (char *filename, script_t * scriptbuf)
{
  int scriptfp;
  struct stat filestat;

  scriptfp = open (filename, O_NONBLOCK + O_RDONLY);
  if (scriptfp == -1)
    {
      return (-1);
    }
  fstat (scriptfp, &filestat);
  scriptbuf->buffer =
    realloc (scriptbuf->buffer, scriptbuf->size + filestat.st_size);
  if (scriptbuf->buffer == NULL)
    {
      return (-1);
    }
/* make room */
  memmove (scriptbuf->buffer + scriptbuf->curpos + filestat.st_size,
	   scriptbuf->buffer + scriptbuf->curpos + 1,
	   scriptbuf->size - scriptbuf->curpos);
  scriptbuf->size +=
    read (scriptfp, scriptbuf->buffer + scriptbuf->curpos, filestat.st_size);
  close (scriptfp);
  return (0);
}



/*	
 * Loadscript attempts to open the script and put it in the script buffer
 * if filename is NULL, stdin is read
 * Returns 0 on success, -1 on error 
 */
int
loadscript (char *filename, script_t * scriptbuf)
{
  int scriptfp;
  struct stat filestat;
  char *here;

  scriptfp = open (filename, O_NONBLOCK + O_RDONLY);
  if (scriptfp == -1)
    {				/* open failed */
      return -1;
    }

  fstat (scriptfp, &filestat);
  scriptbuf->name = (char *) calloc (sizeof (char), strlen (filename) + 1);
  if (scriptbuf->name)
    {
      memcpy (scriptbuf->name, filename, strlen (filename));
    }

  scriptbuf->size = filestat.st_size;
  scriptbuf->mtime = filestat.st_mtime;
  scriptbuf->uid = filestat.st_uid;
  scriptbuf->gid = filestat.st_gid;

  scriptbuf->buffer = (char *) calloc (sizeof (char), filestat.st_size + 2);
  read (scriptfp, scriptbuf->buffer, filestat.st_size);
  scriptbuf->curpos = 0;

  /* skip over the first line, if it starts #! 
   * This means that offset will start at the beginning of 
   * the second line in most cases.  
   */

  here = scriptbuf->buffer;
  if (memcmp (scriptbuf->buffer, "#!", 2) == 0)
    {
      global.bangscript = TRUE;
      while ((scriptbuf->curpos < scriptbuf->size) &&
	     ((char) (here[scriptbuf->curpos]) != '\n'))
	{
	  (scriptbuf->curpos)++;
	}
      (scriptbuf->curpos)++;
    }

  close (scriptfp);
  return (0);
}


/*
 * Freescript frees the memory used by a script structure 
 * Returns 0 on success, -1 on error 
 */
int
freescript (script_t * scriptbuf)
{
  free (scriptbuf->name);
  free (scriptbuf->buffer);
  return (0);
}

/*
 * PrintParseError ( char * error, int line number )
 *
 *
 */

void
PrintParseError (char *error, int linenum)
{
  printf ("HTTP/1.0 500 Server Error\n"
	  "Content-Type: text/html\n\n"
	  "<html><body><b><font color=#CC0000>" PACKAGE_NAME
	  " CGI Error</font></b><br>\n" "Error: %s", error);
  if (linenum > 0)
    {
      printf (" near line %i\n", linenum);
    }
  printf ("</body></html>\n");
}

/*
 *
 * scanline ( char *buf, char *end)
 *  given a pointer to a buffer, cound the newline characters between
 *  the buf and end
 */

int
scanline (char *buf, char *end)
{
  int count = 0;
  while (buf < end)
    {
      if (*buf == '\n')
	{
	  count++;
	}
      buf++;
    }

  return (count + 1);
}


/*
 * PushTokenChain - Adds a token to the end of the token chain 
   returun TRUE on succes; FALSE on failure.
 */

int
PushTokenChain (enum tag_t tag, char *buf, size_t len, int depth)
{
  token_t *cur = token_list;
  token_t /*@null@ */  * prev = NULL;

  /* Exception:  We don't store null sized html */
  if ((tag == HTML) && (len == 0))
    {
      return (TRUE);
    }

  /* Exception: We don't store whitespace html if not in verbatim mode */
  if ((tag == HTML) && (global.verbatim == FALSE) &&
      (strspn (buf, " \t\n\r") == strlen (buf)))
    {
      return (TRUE);
    }

  /* walk the chain */
  while (cur != NULL)
    {
      prev = cur;
      cur = (token_t *) cur->next;
    }

  cur = malloc (sizeof (token_t) + len + 1);
  if (cur)
    {
      if (token_list == NULL)
	{
	  token_list = cur;
	}
      if (prev)
	{
	  prev->next = (void *) cur;
	}
      cur->tag = tag;
      memcpy (cur->buf, buf, len);
      cur->len = len;
      cur->depth = depth;
      cur->next = NULL;
      return (TRUE);
    }
  else
    {
      return (FALSE);
    }

}


/* 
 * FindTag - given a char buffer, len, and tag to find return the offset
 * for the tag, or -1 if not found 
 */

int
FindTag (char *buf, size_t len, char *tag)
{
  size_t offset = 0;

  while ((memcmp (buf + offset, tag, 2) != 0) && (offset <= len))
    {
      offset++;
    }

  if (offset > len)
    {
      return -1;
    }
  else
    {
      return offset;
    }
}



/*
 * buildTokenList 
 * This is a dumb tokenizer, takes a script and pushes tokens to the
 * the given list
 * returns 	0 if all ok
 * 		-1 if some kind of problem
 */
int
buildTokenList (script_t * script)
{
  int state = HTML;
  char *buf, *cp, *start, *end, *tmp;
  size_t linenum;
  char errstr[200];
  char **argv;


  size_t depth = 0;
  int seenelse = FALSE;

  int err, e_token, e_token2;


  cp = script->buffer + script->curpos;

  err = 0;
  e_token = NO_ERROR;
  e_token2 = NO_ERROR;

  while (*cp)
    {
      buf = script->buffer;
      start = strstr (cp, "<?");
      end = strstr (cp, "?>");

      if (start && !end)
	{
	  err = MISSING_X;
	  e_token = GT;
	}
      if ((err == NO_ERROR) && (end && !start))
	{
	  err = MISSING_X;
	  e_token = LT;
	}
      if ((err == NO_ERROR) && (end < start))
	{
	  err = MISSING_X;
	  e_token = LT;
	}
      if ((err == NO_ERROR) && (start))
	{
	  tmp = strstr (start + 1, "<?");
	  if (tmp && (tmp < end))
	    {
	      err = MISSING_X;
	      e_token = GT;
	    }
	}

      if (err != NO_ERROR)
	{
	  snprintf (errstr, 200, global_error[err], global_error[e_token],
		    scanline (buf, cp));
	  PrintParseError (errstr, 0);
	  return (-1);
	}

      if (state == HTML)
	{
	  if (start == NULL)
	    {
	      PushTokenChain (HTML, cp, strlen (cp) + 1, depth);
	      cp += strlen (cp);
	    }
	  else
	    {
	      *start = '\0';
	      tmp = cp;
	      while ((tmp = index (tmp, '\n')) != NULL)
		{
		  linenum++;
		  tmp++;
		}
	      PushTokenChain (HTML, cp, strlen (cp) + 1, depth);
	      state = NOOP;
	      cp = start;
	    }
	}

      if (state != HTML)
	{
	  cp += 2;		/* skip over the <? */

	  if (strchr ("\n\r\t ", (int) *cp))
	    state = RUN;
	  if (memcmp (cp, global_tag[IF], 2) == 0)
	    state = IF;
	  if (memcmp (cp, global_tag[ELSE], 2) == 0)
	    state = ELSE;
	  if (memcmp (cp, global_tag[FI], 2) == 0)
	    state = FI;
	  if (memcmp (cp, global_tag[ABORT], 2) == 0)
	    state = ABORT;
	  if (memcmp (cp, global_tag[INCLUDE], 2) == 0)
	    {
	      state = INCLUDE;
	    }


	  while ((!strchr ("\n\r\t ", (int) *(cp))) && (cp < end))
	    cp++;		/* skip over any junk past the tag */

	  /* we should be at the beginning of the tag now */
	  *end = '\0';		/* make it a string */

	  tmp = cp;
	  while ((tmp = index (tmp, '\n')) != NULL)
	    {
	      linenum++;
	      tmp++;
	    }


	  /* Do we have problems? */

	  if (((state == FI) || (state == ELSE)) && (depth <= 0))
	    {
	      err = X_BEFORE_Y;
	      e_token = state;
	      e_token2 = IF;
	    }


	  if (state == ELSE)
	    {
	      if (seenelse == TRUE)
		{
		  err = X_BEFORE_Y;
		  e_token = ELSE;
		  e_token2 = FI;
		}
	      else
		{
		  seenelse = TRUE;
		  depth--;
		}
	    }

	  if (state == FI)
	    {
	      depth--;
	      seenelse = FALSE;
	    }

	  if (err != NO_ERROR)
	    {
	      snprintf (errstr, 200, global_error[err],
			global_tag[e_token], global_tag[e_token2],
			scanline (buf, cp));
	      PrintParseError (errstr, 0);
	      return (-1);
	    }

	  PushTokenChain (state, cp, strlen (cp) + 1, depth);

	  if (state == IF)
	    {
	      depth++;
	      seenelse = FALSE;
	    }

	  if (state == ELSE)
	    {
	      depth++;
	    }

	  /* Handle include now */
	  if (state == INCLUDE)
	    {
	      script->curpos = (end - buf) + 2;
	      e_token = scanline (buf, cp);
	      argc_argv (cp, &argv, "");
	      if (insertScript (argv[0], script) == -1)
		{
		  snprintf (errstr, 200, global_error[OPENFAIL],
			    argv[0], e_token);
		  PrintParseError (errstr, 0);
		  return (-1);
		}
	      free (argv);
	      end = script->buffer + script->curpos - 2;
	    }

	  cp = end + 2;		/* and move me on to the endtag */
	  state = HTML;		/* and reset state to HTML      */
	}
    }


  return (0);
}



/* 
 * FindHeadBlock - resizes the script so that it only contains
 *  the header block
 */
int
FindHeadBlock (script_t * script)
{
  char *offset = NULL;
  char *cp = script->buffer;
  size_t size;


  if (!(offset = strstr (cp, "\n\n")))
    {
      offset = strstr (cp, "\r\n\r\n");
    }

  if (!offset)
    {
      return (-1);		/* there was no head block */
    }
  else
    {
      size = offset - cp + 2;
      script->buffer = realloc (script->buffer, size);
      script->size = size;
      return (0);
    }
}


/* ProcessTokenChain ( token_t *chain )
 * Processes the token chain, starting at the token passed
 * exits when the next token is of a lower depth than the
 * token it was given.
 */

int
ProcessTokenChain (token_t * token)
{
  int depth;
  int retval;

  depth = token->depth;
  while ((token != NULL) && (token->depth >= depth))
    {
      // fprintf (stderr, "[%d] [%d] [%s]\n", token->depth, token->tag, token->buf);	    
      /* html */
      if (token->tag == HTML)
	{
	  write (STDOUT, token->buf, strlen(token->buf));
	}

      /* abort */
      if (token->tag == ABORT)
	{
	  /* set all the tokens below me to No-op */
	  while (token != NULL)
	    {
	      token->tag = NOOP;
	      token = (void *) token->next;
	    }
	  return (-1);
	}

      /* run */
      if (token->tag == RUN)
	{
	  ExecSubshell (token->buf);
	}

      /* Note - We use 0 to mean TRUE, and !0 to mean FALSE   
       * That's backwards from coding logic, but the way that
       * most programs work.
       */

      /* if  */
      if (token->tag == IF)
	{
	  retval = ExecSubshell (token->buf);
	  if (!retval)
	    {			/* TRUE */
	      if (ProcessTokenChain ((token_t *) token->next) == -1)
		{
		  return (-1);
		}
	      token = (void *) token->next;
	      while ((token != NULL) && (token->depth > depth))
		{
		  token = (void *) token->next;
		  if ((token->tag == ELSE) && (token->depth == depth))
		    {
		      token = (void *) token->next;
		    }
		}
	    }			/* this will move us to the fi */
	  if (retval)
	    {			/* FALSE */
	      token = (void *) token->next;
	      while ((token != NULL) && (token->depth > depth))
		{
		  token = (void *) token->next;
		  if ((token->tag == ELSE) && (token->depth == depth))
		    {
		      if (ProcessTokenChain ((token_t *) token->next) ==
			  -1)
			{
			  return (-1);
			}
		      token = (void *) token->next;
		    }
		}
	    }			/* this moves us to the fi */
	}			/* end if */
      token = (void *) token->next;
    }

  return (0);
}

int
PrintTokenChain (token_t * token)
{

  const char *tags[] =
    { "HTML", "RUN", "IF", "ELSE", "FI", "ABORT", "INCLUDE", "NOOP" };

  while (token != NULL)
    {
      printf ("Token: type %6s, size %3d at depth %2d--%s--\n",
	      tags[token->tag], token->len, token->depth, token->buf);
      token = (void *) token->next;
    }

  return 0;
}

void
sig_handler (int signo)
{
  switch (signo)
    {
    case SIGCHLD:
      exit (-1);
      break;
    default:
      break;
    }
}

/* 
 * Attempt to open a subshell, return TRUE if it worked 
 */

int
OpenSubshell (void)
{
  int retcode = 0;
  char **argv;

  /* Pipe fd's are defined in haserl.h */

  retcode = pipe (&global_subshell_pipe[PARENT_IN]);
  if (retcode == 0)
    {
      retcode = pipe (&global_subshell_pipe[PARENT_CTRLIN]);
    }
  if (retcode == 0)
    {
      signal (SIGCHLD, sig_handler);	/* die if the child dies early */
      global_subshell_pid = fork ();
      if (global_subshell_pid == -1)
	{
	  return FALSE;
	}

      if (global_subshell_pid == 0)
	{
	  /* I'm the child */
	  signal (SIGCHLD, SIG_IGN);
	  dup2 (global_subshell_pipe[PARENT_IN], STDIN);
	  close (global_subshell_pipe[PARENT_IN]);
	  close (global_subshell_pipe[PARENT_OUT]);	/* we won't write on this pipe */
	  if (global_subshell_pipe[CHILD_CTRLOUT] != 5)
	    {
	      /* force the outbound control pipe to be fd 5 */
	      dup2 (global_subshell_pipe[CHILD_CTRLOUT], 5);
	      close (global_subshell_pipe[CHILD_CTRLOUT]);
	    }
	  argc_argv (global.shell, &argv, "");
	  execv (argv[0], argv);
	  free (argv);


	  /* if we get here, we had a failure, so close down all the pipes */
	  PrintParseError ("Unable to start subshell - critical error.", 0);
	  exit (-1);
	}
      else
	{
	  /* I'm parent, move along please */
	}
    }

  /* control should get to this point only in the parent.
   */
  return (0);
}


int
CloseSubshell (void)
{
  int x;

  signal (SIGCHLD, SIG_IGN);

  /* closing the pipe should close the subshell if it is a bash-like shell */
  for (x = 0; x < 4; x++)
    {
      close (global_subshell_pipe[x]);
    }

  /* but kill it just to make sure */
  kill (global_subshell_pid, 9);

  return TRUE;
}


int
ExecSubshell (char *buf)
{
  char *mybuf;
  const char postfix[] = "\necho $? >&5\n";
  int count = 0;
  int retval;

  mybuf = calloc (sizeof (char), strlen (buf) + strlen (postfix) + 1);
  if (mybuf == NULL)
    {
      return (-1);
    }
  memcpy (mybuf, buf, strlen (buf));
  memcpy (mybuf + strlen (buf), postfix, strlen (postfix));
  write (global_subshell_pipe[PARENT_OUT], mybuf, strlen (mybuf));

  /* wait for the result to come back */
  /* clear mybuf to start */
  memset (mybuf, 0, sizeof (mybuf));
  while (mybuf[count] != '\n')
    {
      read (global_subshell_pipe[PARENT_CTRLIN], &mybuf[count], 1);
      if (mybuf[count] != '\n')
	{
	  count++;
	}
    }

  mybuf[count] = 0;
  retval = atoi (mybuf);
  free (mybuf);
  return (retval);
}


int
parseCommandLine (int argc, char *argv[])
{
  int c;
  int option_index = 0;

  /* set optopt and optind to 0 to reset getopt_long -
   * we may call it multiple times
   */
  optopt = 0;
  optind = 0;

  while ((c = getopt_long (argc, argv, gs_short_options,
			   ga_long_options, &option_index)) != -1)
    {
      switch (c)
	{
	case 'D':
	  global.debug = TRUE;
	  break;
	case 's':
	  global.shell = optarg;
	  break;
	case 'u':
	  if (optarg)
	    {
	      global.uploadkb = atoi (optarg);
	    }
	  else
	    {
	      global.uploadkb = MAX_UPLOAD_KB;
	    }
	  break;
	case 'a':
	  global.acceptall = TRUE;
	  break;
	case 'V':
	  global.verbatim = TRUE;
	  break;
	case 'U':
	  global.uploaddir = optarg;
	  break;
	case 'v':
	case 'h':
	  printf ("This is " PACKAGE_NAME " version " PACKAGE_VERSION ""
		  "(http://haserl.sourceforge.net)\n");
	  exit (0);
	  break;
	}
    }
  return (optind);
}

int
BecomeUser (uid_t uid, gid_t gid)
{
  /* This silently fails if it doesn't work */
  setgid (gid);
  setuid (uid);
  return (0);
}

/*
 * Assign default values to the global structure
 */

void
assignGlobalStartupValues ()
{
  global.uploadkb = 0;		/* how big an upload do we allow (0 for none)   */
  global.shell = SUBSHELL_CMD;	/* The shell we use                             */
  global.uploaddir = TEMPDIR;	/* where to upload to                           */
  global.verbatim = FALSE;	/* Don't print whitespace html segments         */
  global.debug = FALSE;		/* Not int debug mode.                          */
  global.bangscript = FALSE;	/* don't read environ for postvars, etc         */
  global.acceptall = FALSE;	/* don't allow POST data for GET method         */

}

/*-------------------------------------------------------------------------
 *
 * Main 
 *
 *------------------------------------------------------------------------*/

int
main (int argc, char *argv[])
{

  script_t script;
  int retval = 0;
  char *filename = NULL;

  char **av = NULL;
  int command;

  assignGlobalStartupValues ();

  /* argv has multiple elements, split it (we were called as a #! script */
  if ((argc > 1) && (index (argv[1], ' ') != NULL))
    {
      command = argc_argv (argv[1], &av, "");
      /* and we need to add the command for optarg */
      av = (char **) realloc (av, sizeof (char *) * (command + 2));
      if (av)
	{
	  memmove (&av[1], &av[0], (command) * (sizeof (char *)));
	  av[0] = argv[0];
	  parseCommandLine (command + 1, av);
	  argv[1] = av[1];
	  free (av);
	}
      filename = argv[2];
    }

  command = parseCommandLine (argc, argv);

  if ((command < argc) && (filename == NULL))
    {
      filename = argv[command];
    }

  if (filename == NULL)
    {
      /* we were run, instead of called as a shell script */
      puts ("This is " PACKAGE_NAME " version " PACKAGE_VERSION "\n"
	    "This program runs as a cgi interpeter, not interactively\n"
	    "Please see:  http://haserl.sourceforge.net");
      return (0);
    }


  if (loadscript (filename, &script))
    {
      return (-1);
    }


/* drop permissions */
  BecomeUser (script.uid, script.gid);

/* Read the current environment into our chain */
  readenv ();
  sessionid ();
  wcversion ();
  haserlflags ();
  CookieVars ();


  /* If we have a request method, and we were run as a #! style script */
  if (getenv ("REQUEST_METHOD") && (global.bangscript == TRUE))
    {
      if (strcasecmp (getenv ("REQUEST_METHOD"), "GET") == 0)
	{
	  if (global.acceptall == TRUE)
	    {
	      retval = ReadCGIPOSTValues ();
	    }
	  if (retval == 0)
	    {
	      retval = ReadCGIQueryString ();
	    }
	}

      if (strcasecmp (getenv ("REQUEST_METHOD"), "POST") == 0)
	{
	  if (global.acceptall == TRUE)
	    {
	      retval = ReadCGIQueryString ();
	    }
	  retval = ReadCGIPOSTValues ();
	}

      if (strcasecmp (getenv ("REQUEST_METHOD"), "HEAD") == 0)
	{
	  retval = FindHeadBlock (&script);
	}
    }

  if (retval == 0)
    {
      retval = OpenSubshell ();
    }

  if ((retval == 0) && (buildTokenList (&script) == 0))

    freescript (&script);

  if (retval == 0)
    {
      if (global.debug == TRUE)
	{
	  PrintTokenChain ((void *) token_list);
	}
      else
	{
	  retval = ProcessTokenChain ((void *) token_list);
	  CloseSubshell ();
	}
    }

  freeenvchain ();
  FreeTokenChain ();
  return (retval);
}
